"""Generate CSS rules for Excel in-cell rich-value images."""

from io import BytesIO
import logging
from zipfile import ZipFile
from PIL import Image, UnidentifiedImageError
from xlsx2html.utils.image import bytes_to_datauri

from xx2html.core.types import CellDimensions


def get_incell_css(
    vm_ids: set[str],
    vm_ids_dimension_references: dict[str, CellDimensions],
    vm_cell_vm_ids: dict[str, str],
    incell_images_refs: dict[str | None, str],
    archive: ZipFile,
) -> str:
    """Create per-vm and per-cell CSS rules for in-cell image rendering."""
    styles = []
    archive_namelist = archive.namelist()
    image_specs: dict[str, dict[str, object]] = {}

    for vm_id in sorted(vm_ids):
        rel_id = f"rId{vm_id}"

        # logging.debug(f"Transform (wb|incell): Working with vm(rId): {rel_id}")
        logging.info(f"Transform (wb|incell): Working with vm(rId): {rel_id}")
        # if rel_id in incell_images:
        # if vm_id in incell_images_refs:
        target_path = incell_images_refs.get(vm_id, None)
        if target_path is None:
            logging.warning(
                f"get_incell_css: vm(rId) -> {rel_id} not found in incell images references!"
            )
            continue
        if target_path not in archive_namelist:
            logging.warning(
                f"get_incell_css: vm(rId) -> image '{target_path}' not found in excel file!"
            )
            continue
        # logging.debug(f"Transform (wb|incell): Reading vm(rId): {rel_id} -> file at {target_path}")
        logging.info(f"get_incell_css: Reading vm(rId): {rel_id} -> file '{target_path}'")
        with archive.open(target_path) as ifile:
            image_bytes = ifile.read()

        src = bytes_to_datauri(BytesIO(image_bytes), target_path)
        width = None
        height = None
        try:
            with Image.open(BytesIO(image_bytes)) as image:
                width, height = image.size
        except (UnidentifiedImageError, OSError) as image_exc:
            logging.warning(
                "get_incell_css: Unable to read image size for vm(rId)=%s (%s): %r",
                rel_id,
                target_path,
                image_exc,
            )
        image_specs[vm_id] = {"src": src, "width": width, "height": height}

        styles.append(
            """
.vm-richvaluerel_rid%s img {
    content:url("%s");
    display:block;
}
                """
            % (vm_id, src)
        )

    for cell_class_name, cell_dimensions in vm_ids_dimension_references.items():
        vm_id = vm_cell_vm_ids.get(cell_class_name)
        image_spec = image_specs.get(vm_id, {})
        cell_width = cell_dimensions.get("width", 0)
        cell_height = cell_dimensions.get("height", 0)
        img_width = image_spec.get("width")
        img_height = image_spec.get("height")

        width_rule = "width: 100%;"
        height_rule = "height: 100%;"
        aspect_ratio_rule = ""

        if (
            isinstance(cell_width, int)
            and isinstance(cell_height, int)
            and cell_width > 0
            and cell_height > 0
            and isinstance(img_width, int)
            and isinstance(img_height, int)
            and img_width > 0
            and img_height > 0
        ):
            cell_aspect_ratio = cell_width / cell_height
            image_aspect_ratio = img_width / img_height
            if image_aspect_ratio >= cell_aspect_ratio:
                width_rule = "width: 100%;"
                height_rule = "height: auto;"
            else:
                width_rule = "width: auto;"
                height_rule = "height: 100%;"
            aspect_ratio_rule = f"aspect-ratio: {img_width} / {img_height};"

        styles.append(
            """
.%s img {
    %s
    %s
    max-width: 100%%;
    max-height: 100%%;
    object-fit: contain;
    %s
}
            """
            % (cell_class_name, width_rule, height_rule, aspect_ratio_rule)
        )

    return "\n".join(styles)



import json
import tempfile
import unittest
from pathlib import Path

from xx2html.cova import initialize


class CovaInitializeTests(unittest.TestCase):
    def test_initialize_returns_merged_config_without_mutating_default(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            config_path = Path(tmp_dir)
            config_file = config_path / "config.json"
            config_file.write_text(json.dumps({"user": True, "a": 2}), encoding="utf-8")

            default_config = {"a": 1, "nested": {"x": 1}}
            system_config = {"sys": True}

            result = initialize(
                app_config_path=str(config_path),
                config_filename="config.json",
                default_config=default_config,
                system_config=system_config,
            )

            self.assertEqual(1, default_config["a"])
            self.assertEqual({"x": 1}, default_config["nested"])

            self.assertEqual(
                {"a": 2, "nested": {"x": 1}, "user": True, "sys": True},
                result,
            )


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

import xx2html.core as core_module
from xx2html import create_xlsx_transform

FIXTURES_DIR = Path(__file__).resolve().parent / "fixtures"

SHEET_HTML = (
    '<section id="{enc_sheet_name}" data-sheet="{sheet_name}">'
    "{table_generated_html}"
    "</section>"
)
SHEETNAME_HTML = '<a class="sheet-nav" href="#{sheet_name}.A1">{sheet_name}</a>'
INDEX_HTML = (
    "<!doctype html><html><head>"
    "{fonts_html}{core_css_html}{user_css_html}{generated_css_html}"
    "{generated_incell_css_html}{conditional_css_html}"
    "</head><body>{sheets_names_generated_html}{sheets_generated_html}{safari_js}</body></html>"
)


def _run_transform(source_file: Path, apply_cf: bool) -> tuple[bool, str | None]:
    transform = create_xlsx_transform(
        sheet_html=SHEET_HTML,
        sheetname_html=SHEETNAME_HTML,
        index_html=INDEX_HTML,
        fonts_html="",
        core_css="",
        user_css="",
        safari_js="",
        apply_cf=apply_cf,
    )
    with tempfile.TemporaryDirectory() as tmp_dir:
        output_file = Path(tmp_dir) / "output.html"
        return transform(str(source_file), str(output_file), "en_US")


class ConditionalFormattingPipelineTests(unittest.TestCase):
    def test_apply_cf_true_calls_condif2css_processor(self):
        source_file = FIXTURES_DIR / "merged_cells_cf.xlsx"

        with patch(
            "xx2html.core.process_conditional_formatting",
            wraps=core_module.process_conditional_formatting,
        ) as process_mock:
            ok, err = _run_transform(source_file, apply_cf=True)

        self.assertTrue(ok, err)
        self.assertGreaterEqual(process_mock.call_count, 1)
        args, kwargs = process_mock.call_args
        self.assertEqual("Data", args[0].title)
        self.assertTrue(kwargs.get("fail_ok", False))

    def test_apply_cf_false_skips_condif2css_processor(self):
        source_file = FIXTURES_DIR / "merged_cells_cf.xlsx"

        with patch(
            "xx2html.core.process_conditional_formatting",
            wraps=core_module.process_conditional_formatting,
        ) as process_mock:
            ok, err = _run_transform(source_file, apply_cf=False)

        self.assertTrue(ok, err)
        self.assertEqual(0, process_mock.call_count)


if __name__ == "__main__":
    unittest.main()



import unittest

from openpyxl import Workbook
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side

from xx2html.core.css import (
    CssRegistry,
    create_get_css_components_from_cell,
    get_border_classes_from_cell,
)


class CssRegistryTests(unittest.TestCase):
    @staticmethod
    def _color_resolver(color):
        rgb = getattr(color, "rgb", None)
        return rgb if isinstance(rgb, str) else None

    def test_register_font_underline_uses_text_decoration(self):
        registry = CssRegistry(get_css_color=lambda _: None)

        class_name = registry.register_font_underline()

        self.assertEqual("xlsx_cell_font_underline", class_name)
        self.assertEqual("text-decoration: underline;", registry.classes[class_name])

    def test_default_classes_dict_is_not_shared_between_instances(self):
        registry_a = CssRegistry(get_css_color=lambda _: None)
        registry_b = CssRegistry(get_css_color=lambda _: None)

        registry_a.register_font_bold()

        self.assertIn("xlsx_cell_font_bold", registry_a.classes)
        self.assertNotIn("xlsx_cell_font_bold", registry_b.classes)

    def test_register_core_class_helpers(self):
        registry = CssRegistry(get_css_color=self._color_resolver)

        self.assertEqual("xlsx_cell_font_size_12", registry.register_font_size(12))
        self.assertEqual("xlsx_element_height_33", registry.register_height(33))
        self.assertEqual(
            "xlsx_cell_font_bold", registry.register_font_bold()
        )
        self.assertEqual(
            "xlsx_cell_font_italic", registry.register_font_italic()
        )
        self.assertEqual(
            "xlsx_cell_font_color_FF112233",
            registry.register_font_color(Side(color="FF112233").color),
        )
        self.assertEqual(
            "xlsx_cell_background_color_FF010203",
            registry.register_background_color(Side(color="FF010203").color),
        )

        self.assertIn("font-size: 12px;", registry.classes.values())
        self.assertIn("height: 33px;", registry.classes.values())
        self.assertIn("font-weight: bold;", registry.classes.values())
        self.assertIn("font-style: italic;", registry.classes.values())

    def test_register_color_methods_return_none_when_color_unresolved(self):
        registry = CssRegistry(get_css_color=lambda _: None)
        color = Side(color="FF112233").color

        self.assertIsNone(registry.register_font_color(color))
        self.assertIsNone(registry.register_background_color(color))

    def test_register_border_uses_default_style_and_optional_color(self):
        registry = CssRegistry(get_css_color=self._color_resolver)
        color = Side(color="FFAABBCC").color

        class_name = registry.register_border("dashDot", "left", color)

        self.assertEqual("border_dashDot_l_FFAABBCC", class_name)
        class_value = registry.classes[class_name]
        self.assertIn("border-left-style: solid;", class_value)
        self.assertIn("border-left-width: 1px;", class_value)
        self.assertIn("border-left-color:", class_value)

    def test_register_border_returns_none_when_style_none(self):
        registry = CssRegistry(get_css_color=self._color_resolver)
        color = Side(color="FF000000").color
        self.assertIsNone(registry.register_border(None, "top", color))

    def test_get_border_classes_from_cell_reads_all_sides(self):
        wb = Workbook()
        ws = wb.active
        cell = ws["A1"]
        cell.border = Border(
            left=Side(style="thin", color="FF0000FF"),
            right=Side(style="dashed", color="FF00FF00"),
            top=Side(style="double", color="FFFF0000"),
            bottom=Side(style="medium", color="FF123456"),
        )
        registry = CssRegistry(get_css_color=self._color_resolver)

        classes = get_border_classes_from_cell(cell, registry)

        self.assertEqual(4, len(classes))
        self.assertTrue(any(cls.startswith("border_thin_l_") for cls in classes))
        self.assertTrue(any(cls.startswith("border_dashed_r_") for cls in classes))
        self.assertTrue(any(cls.startswith("border_double_t_") for cls in classes))
        self.assertTrue(any(cls.startswith("border_medium_b_") for cls in classes))

    def test_get_css_components_from_cell_collects_alignment_fill_font_and_merged_borders(self):
        wb = Workbook()
        ws = wb.active
        cell = ws["A1"]
        merged_neighbor = ws["B1"]

        cell.alignment = Alignment(horizontal="center", vertical="top")
        cell.fill = PatternFill(fill_type="solid", fgColor="FF00FF00")
        cell.font = Font(
            sz=14,
            color="FF112233",
            b=True,
            i=True,
            u="single",
        )
        cell.border = Border(left=Side(style="thin", color="FF0000FF"))
        merged_neighbor.border = Border(right=Side(style="thick", color="FF101010"))

        registry = CssRegistry(get_css_color=self._color_resolver)
        get_components = create_get_css_components_from_cell(registry)

        styles, classes = get_components(cell, merged_cell_map={"cells": [merged_neighbor]})

        self.assertEqual("center", styles["text-align"])
        self.assertEqual("top", styles["vertical-align"])
        self.assertIn("xlsx_cell_font_size_14", classes)
        self.assertIn("xlsx_cell_font_bold", classes)
        self.assertIn("xlsx_cell_font_italic", classes)
        self.assertIn("xlsx_cell_font_underline", classes)
        self.assertIn("xlsx_cell_font_color_FF112233", classes)
        self.assertIn("xlsx_cell_background_color_FF00FF00", classes)
        self.assertTrue(any(cls.startswith("border_thin_l_") for cls in classes))
        self.assertTrue(any(cls.startswith("border_thick_r_") for cls in classes))

    def test_get_css_components_from_cell_warns_on_unsupported_pattern_type(self):
        wb = Workbook()
        ws = wb.active
        cell = ws["A1"]
        cell.fill = PatternFill(fill_type="darkGrid", fgColor="FF00FF00")

        registry = CssRegistry(get_css_color=self._color_resolver)
        get_components = create_get_css_components_from_cell(registry)

        with self.assertLogs(level="WARNING") as log_ctx:
            styles, classes = get_components(cell)

        self.assertEqual({}, styles)
        self.assertIsInstance(classes, set)
        self.assertTrue(
            any("Pattern type is not supported: darkGrid" in msg for msg in log_ctx.output)
        )


if __name__ == "__main__":
    unittest.main()



import tempfile
import unittest
from pathlib import Path

from xx2html import create_xlsx_transform

FIXTURES_DIR = Path(__file__).resolve().parent / "fixtures"

SHEET_HTML = (
    '<section id="{enc_sheet_name}" data-sheet="{sheet_name}">'
    "{table_generated_html}"
    "</section>"
)
SHEETNAME_HTML = '<a class="sheet-nav" href="#{sheet_name}.A1">{sheet_name}</a>'
INDEX_HTML = (
    "<!doctype html><html><head>"
    "{fonts_html}{core_css_html}{user_css_html}{generated_css_html}"
    "{generated_incell_css_html}{conditional_css_html}"
    "</head><body>{sheets_names_generated_html}{sheets_generated_html}{safari_js}</body></html>"
)


def _render_fixture(source_file: Path, apply_cf: bool) -> str:
    transform = create_xlsx_transform(
        sheet_html=SHEET_HTML,
        sheetname_html=SHEETNAME_HTML,
        index_html=INDEX_HTML,
        fonts_html="",
        core_css="",
        user_css="",
        safari_js="",
        apply_cf=apply_cf,
    )

    with tempfile.TemporaryDirectory() as tmp_dir:
        output_file = Path(tmp_dir) / "output.html"
        ok, err = transform(str(source_file), str(output_file), "en_US")
        if not ok:
            raise AssertionError(f"Transform failed for {source_file.name}: {err}")
        return output_file.read_text(encoding="utf-8")


class FixturesIntegrationTests(unittest.TestCase):
    def test_merged_cells_and_conditional_formatting_fixture(self):
        html = _render_fixture(FIXTURES_DIR / "merged_cells_cf.xlsx", apply_cf=True)

        self.assertIn('colspan="2"', html)
        self.assertIn('rowspan="2"', html)
        self.assertIn("/*conditional formatting*/", html)
        self.assertIn("Data!C1", html)

    def test_incell_image_fixture(self):
        html = _render_fixture(FIXTURES_DIR / "incell_image.xlsx", apply_cf=False)

        self.assertIn("vm-richvaluerel_rid1", html)
        self.assertIn("incell-image", html)
        self.assertIn("content:url(\"data:image/png;base64", html)
        self.assertIn("object-fit: contain;", html)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from pathlib import Path
import tempfile
from zipfile import ZIP_DEFLATED, ZipFile
from xml.etree import ElementTree as ET

from openpyxl import Workbook
from openpyxl.formatting.rule import FormulaRule
from openpyxl.styles import PatternFill
from PIL import Image

ROOT_DIR = Path(__file__).resolve().parents[2]
FIXTURES_DIR = ROOT_DIR / "tests" / "fixtures"

MAIN_NS = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
PKG_REL_NS = "http://schemas.openxmlformats.org/package/2006/relationships"
CONTENT_TYPES_NS = "http://schemas.openxmlformats.org/package/2006/content-types"
RD_NS = "http://schemas.microsoft.com/office/spreadsheetml/2017/richdata"


def _ensure_content_type_entries(content_types_path: Path) -> None:
    ET.register_namespace("", CONTENT_TYPES_NS)
    tree = ET.parse(content_types_path)
    root = tree.getroot()

    def has_default(ext: str) -> bool:
        return root.find(
            f"{{{CONTENT_TYPES_NS}}}Default[@Extension='{ext}']"
        ) is not None

    def has_override(part_name: str) -> bool:
        return root.find(
            f"{{{CONTENT_TYPES_NS}}}Override[@PartName='{part_name}']"
        ) is not None

    if not has_default("png"):
        root.append(
            ET.Element(
                f"{{{CONTENT_TYPES_NS}}}Default",
                {"Extension": "png", "ContentType": "image/png"},
            )
        )

    overrides = {
        "/xl/metadata.xml": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheetMetadata+xml",
        "/xl/richData/rdrichvalue.xml": "application/vnd.ms-excel.rdrichvalue+xml",
        "/xl/richData/rdrichvaluestructure.xml": "application/vnd.ms-excel.rdrichvaluestructure+xml",
        "/xl/richData/richValueRel.xml": "application/vnd.ms-excel.richvaluerel+xml",
    }
    for part_name, content_type in overrides.items():
        if has_override(part_name):
            continue
        root.append(
            ET.Element(
                f"{{{CONTENT_TYPES_NS}}}Override",
                {"PartName": part_name, "ContentType": content_type},
            )
        )

    tree.write(content_types_path, encoding="utf-8", xml_declaration=True)


def _set_vm_attribute_on_a1(sheet_path: Path) -> None:
    ET.register_namespace("", MAIN_NS)
    tree = ET.parse(sheet_path)
    root = tree.getroot()
    sheet_data = root.find(f"{{{MAIN_NS}}}sheetData")
    if sheet_data is None:
        raise RuntimeError("sheetData not found in sheet XML")

    a1_cell = root.find(f".//{{{MAIN_NS}}}c[@r='A1']")
    if a1_cell is None:
        raise RuntimeError("A1 cell not found in sheet XML")

    a1_cell.set("vm", "1")
    tree.write(sheet_path, encoding="utf-8", xml_declaration=True)


def _write_incell_rich_data_parts(tmp_dir: Path) -> None:
    rich_data_dir = tmp_dir / "xl" / "richData"
    rich_data_rels_dir = rich_data_dir / "_rels"
    media_dir = tmp_dir / "xl" / "media"

    rich_data_dir.mkdir(parents=True, exist_ok=True)
    rich_data_rels_dir.mkdir(parents=True, exist_ok=True)
    media_dir.mkdir(parents=True, exist_ok=True)

    metadata_xml = f"""<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<metadata xmlns=\"{MAIN_NS}\">
  <valueMetadata>
    <bk>
      <rc v=\"0\"/>
    </bk>
  </valueMetadata>
</metadata>
"""
    (tmp_dir / "xl" / "metadata.xml").write_text(metadata_xml, encoding="utf-8")

    rich_structure_xml = f"""<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<rd:richValueStructures xmlns:rd=\"{RD_NS}\">
  <rd:s t=\"_localImage\"/>
</rd:richValueStructures>
"""
    (rich_data_dir / "rdrichvaluestructure.xml").write_text(
        rich_structure_xml, encoding="utf-8"
    )

    rich_value_xml = f"""<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<rd:richValueData xmlns:rd=\"{RD_NS}\">
  <rd:rv s=\"0\"><rd:v>0</rd:v><rd:v>inline-image</rd:v></rd:rv>
</rd:richValueData>
"""
    (rich_data_dir / "rdrichvalue.xml").write_text(rich_value_xml, encoding="utf-8")

    rels_xml = f"""<?xml version=\"1.0\" encoding=\"UTF-8\"?>
<Relationships xmlns=\"{PKG_REL_NS}\">
  <Relationship Id=\"rId1\" Type=\"http://schemas.openxmlformats.org/officeDocument/2006/relationships/image\" Target=\"../media/image1.png\"/>
</Relationships>
"""
    (rich_data_rels_dir / "richValueRel.xml.rels").write_text(rels_xml, encoding="utf-8")

    image = Image.new("RGBA", (140, 70), (255, 0, 0, 255))
    image.save(media_dir / "image1.png", format="PNG")


def _zip_dir_to_xlsx(source_dir: Path, output_path: Path) -> None:
    with ZipFile(output_path, "w", compression=ZIP_DEFLATED) as archive:
        for file_path in sorted(source_dir.rglob("*")):
            if file_path.is_dir():
                continue
            archive.write(file_path, file_path.relative_to(source_dir).as_posix())


def create_merged_cells_cf_fixture(output_path: Path) -> None:
    wb = Workbook()
    ws = wb.active
    ws.title = "Data"

    ws["A1"] = "Merged"
    ws.merge_cells("A1:B2")

    ws["C1"] = 10
    ws["C2"] = 2

    warning_fill = PatternFill(start_color="FFFFAA00", end_color="FFFFAA00", fill_type="solid")
    ws.conditional_formatting.add(
        "C1:C2",
        FormulaRule(formula=["TRUE"], fill=warning_fill),
    )

    wb.save(output_path)


def create_incell_image_fixture(output_path: Path) -> None:
    wb = Workbook()
    ws = wb.active
    ws.title = "Images"
    ws["A1"] = "InCell"
    ws.column_dimensions["A"].width = 24
    ws.row_dimensions[1].height = 42
    wb.save(output_path)

    with tempfile.TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)

        with ZipFile(output_path, "r") as archive:
            archive.extractall(tmp_dir)

        _set_vm_attribute_on_a1(tmp_dir / "xl" / "worksheets" / "sheet1.xml")
        _write_incell_rich_data_parts(tmp_dir)
        _ensure_content_type_entries(tmp_dir / "[Content_Types].xml")

        _zip_dir_to_xlsx(tmp_dir, output_path)


def main() -> None:
    FIXTURES_DIR.mkdir(parents=True, exist_ok=True)
    create_merged_cells_cf_fixture(FIXTURES_DIR / "merged_cells_cf.xlsx")
    create_incell_image_fixture(FIXTURES_DIR / "incell_image.xlsx")
    print("Generated fixtures in", FIXTURES_DIR)


if __name__ == "__main__":
    main()



import io
import unittest
from zipfile import ZIP_DEFLATED, ZipFile

from PIL import Image

from xx2html.core.incell import get_incell_css


def _make_png(width: int, height: int) -> bytes:
    image = Image.new("RGBA", (width, height), (255, 0, 0, 255))
    buffer = io.BytesIO()
    image.save(buffer, format="PNG")
    return buffer.getvalue()


class InCellCssTests(unittest.TestCase):
    def test_wide_image_uses_full_width_auto_height(self):
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED) as zf:
            zf.writestr("xl/media/image1.png", _make_png(400, 100))

        zip_buffer.seek(0)
        with ZipFile(zip_buffer, mode="r") as zf:
            css = get_incell_css(
                vm_ids={"1"},
                vm_ids_dimension_references={"cell_1": {"width": 100, "height": 100}},
                vm_cell_vm_ids={"cell_1": "1"},
                incell_images_refs={"1": "xl/media/image1.png"},
                archive=zf,
            )

        self.assertIn(".vm-richvaluerel_rid1 img", css)
        self.assertIn(".cell_1 img", css)
        self.assertIn("width: 100%;", css)
        self.assertIn("height: auto;", css)
        self.assertIn("object-fit: contain;", css)

    def test_tall_image_uses_auto_width_full_height(self):
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED) as zf:
            zf.writestr("xl/media/image2.png", _make_png(100, 400))

        zip_buffer.seek(0)
        with ZipFile(zip_buffer, mode="r") as zf:
            css = get_incell_css(
                vm_ids={"2"},
                vm_ids_dimension_references={"cell_2": {"width": 100, "height": 100}},
                vm_cell_vm_ids={"cell_2": "2"},
                incell_images_refs={"2": "xl/media/image2.png"},
                archive=zf,
            )

        self.assertIn(".vm-richvaluerel_rid2 img", css)
        self.assertIn(".cell_2 img", css)
        self.assertIn("width: auto;", css)
        self.assertIn("height: 100%;", css)
        self.assertIn("object-fit: contain;", css)

    def test_missing_vm_reference_logs_warning_and_skips_vm_content_rule(self):
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED) as zf:
            zf.writestr("xl/media/image1.png", _make_png(50, 50))

        zip_buffer.seek(0)
        with ZipFile(zip_buffer, mode="r") as zf:
            with self.assertLogs(level="WARNING") as log_ctx:
                css = get_incell_css(
                    vm_ids={"1"},
                    vm_ids_dimension_references={"cell_1": {"width": 120, "height": 80}},
                    vm_cell_vm_ids={"cell_1": "1"},
                    incell_images_refs={},
                    archive=zf,
                )

        self.assertTrue(
            any("not found in incell images references" in msg for msg in log_ctx.output)
        )
        self.assertNotIn(".vm-richvaluerel_rid1 img", css)
        self.assertIn(".cell_1 img", css)
        self.assertIn("width: 100%;", css)
        self.assertIn("height: 100%;", css)

    def test_missing_image_file_logs_warning_and_skips_vm_content_rule(self):
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED):
            pass

        zip_buffer.seek(0)
        with ZipFile(zip_buffer, mode="r") as zf:
            with self.assertLogs(level="WARNING") as log_ctx:
                css = get_incell_css(
                    vm_ids={"2"},
                    vm_ids_dimension_references={"cell_2": {"width": 90, "height": 90}},
                    vm_cell_vm_ids={"cell_2": "2"},
                    incell_images_refs={"2": "xl/media/missing.png"},
                    archive=zf,
                )

        self.assertTrue(any("not found in excel file" in msg for msg in log_ctx.output))
        self.assertNotIn(".vm-richvaluerel_rid2 img", css)
        self.assertIn(".cell_2 img", css)

    def test_unreadable_image_falls_back_to_default_fit_rules(self):
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED) as zf:
            zf.writestr("xl/media/image3.png", b"not-a-valid-image")

        zip_buffer.seek(0)
        with ZipFile(zip_buffer, mode="r") as zf:
            with self.assertLogs(level="WARNING") as log_ctx:
                css = get_incell_css(
                    vm_ids={"3"},
                    vm_ids_dimension_references={"cell_3": {"width": 100, "height": 100}},
                    vm_cell_vm_ids={"cell_3": "3"},
                    incell_images_refs={"3": "xl/media/image3.png"},
                    archive=zf,
                )

        self.assertTrue(any("Unable to read image size" in msg for msg in log_ctx.output))
        self.assertIn(".vm-richvaluerel_rid3 img", css)
        self.assertIn(".cell_3 img", css)
        self.assertIn("width: 100%;", css)
        self.assertIn("height: 100%;", css)


if __name__ == "__main__":
    unittest.main()



from xx2html.core import apply_openpyxl_patches, create_xlsx_transform
import xx2html._constants as constants

__version__ = constants.__version__

__all__ = ["__version__", "apply_openpyxl_patches", "create_xlsx_transform"]



import io
import unittest
from unittest.mock import patch
from zipfile import ZIP_DEFLATED, ZipFile

from lxml import etree

import xx2html.core.vm as vm_module
from xx2html.core.vm import _get_xml_from_archive, get_incell_images_refs


class VmTests(unittest.TestCase):
    @staticmethod
    def _valid_rels_xml() -> str:
        return """
<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships">
  <Relationship Id="rId1" Type="http://example" Target="../media/image1.png"/>
</Relationships>
""".strip()

    @staticmethod
    def _valid_structure_xml() -> str:
        return """
<rd:richValueStructures xmlns:rd="http://schemas.microsoft.com/office/spreadsheetml/2017/richdata">
  <rd:s t="_localImage" />
</rd:richValueStructures>
""".strip()

    @staticmethod
    def _valid_richvalue_xml() -> str:
        return """
<rd:richValueData xmlns:rd="http://schemas.microsoft.com/office/spreadsheetml/2017/richdata">
  <rd:rv s="0"><rd:v>0</rd:v><rd:v>ignored</rd:v></rd:rv>
</rd:richValueData>
""".strip()

    @staticmethod
    def _valid_metadata_xml() -> str:
        return """
<metadata xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main">
  <valueMetadata>
    <bk>
      <rc v="0" />
    </bk>
  </valueMetadata>
</metadata>
""".strip()

    def _build_archive(
        self,
        *,
        rels_xml: str | None = None,
        structure_xml: str | None = None,
        richvalue_xml: str | None = None,
        metadata_xml: str | None = None,
    ) -> io.BytesIO:
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED) as zf:
            if rels_xml is not None:
                zf.writestr("xl/richData/_rels/richValueRel.xml.rels", rels_xml)
            if structure_xml is not None:
                zf.writestr("xl/richData/rdrichvaluestructure.xml", structure_xml)
            if richvalue_xml is not None:
                zf.writestr("xl/richData/rdrichvalue.xml", richvalue_xml)
            if metadata_xml is not None:
                zf.writestr("xl/metadata.xml", metadata_xml)
        zip_buffer.seek(0)
        return zip_buffer

    def test_get_incell_images_refs_extracts_vm_to_target_map(self):
        zip_buffer = self._build_archive(
            rels_xml=self._valid_rels_xml(),
            structure_xml=self._valid_structure_xml(),
            richvalue_xml=self._valid_richvalue_xml(),
            metadata_xml=self._valid_metadata_xml(),
        )
        with ZipFile(zip_buffer, mode="r") as zf:
            refs, err = get_incell_images_refs(zf)

        self.assertIsNone(err)
        self.assertIn("1", refs)
        self.assertTrue(refs["1"].endswith("/image1.png"))

    def test_try_parse_int_handles_none_and_invalid_values(self):
        self.assertIsNone(vm_module._try_parse_int(None))
        self.assertIsNone(vm_module._try_parse_int("x"))
        self.assertEqual(7, vm_module._try_parse_int("7"))

    def test_get_xml_from_archive_handles_missing_file(self):
        zip_buffer = self._build_archive()
        with ZipFile(zip_buffer, mode="r") as zf:
            root, err = _get_xml_from_archive(zf, "does/not/exist.xml")
        self.assertIsNone(root)
        self.assertIsInstance(err, KeyError)

    def test_get_xml_from_archive_handles_invalid_xml(self):
        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, mode="w", compression=ZIP_DEFLATED) as zf:
            zf.writestr("broken.xml", "<x>")
        zip_buffer.seek(0)
        with ZipFile(zip_buffer, mode="r") as zf:
            root, err = _get_xml_from_archive(zf, "broken.xml")
        self.assertIsNone(root)
        self.assertIsInstance(err, etree.XMLSyntaxError)

    def test_get_xml_from_archive_handles_unexpected_exception(self):
        class BrokenArchive:
            def open(self, *_args, **_kwargs):  # noqa: ANN002, ANN003
                raise RuntimeError("boom")

        root, err = _get_xml_from_archive(BrokenArchive(), "x.xml")  # type: ignore[arg-type]
        self.assertIsNone(root)
        self.assertIsInstance(err, RuntimeError)

    def test_get_local_image_type_indexes_filters_expected_types(self):
        structure_tree = etree.fromstring(
            """
<rd:richValueStructures xmlns:rd="http://schemas.microsoft.com/office/spreadsheetml/2017/richdata">
  <rd:s t="_localImage" />
  <rd:s t="otherType" />
  <rd:s t="_localImage" />
</rd:richValueStructures>
""".strip().encode("utf-8")
        )
        self.assertEqual(
            {"0", "2"},
            vm_module._get_local_image_type_indexes(structure_tree),
        )

    def test_get_rich_data_value_targets_skips_invalid_and_keeps_valid_records(self):
        richvalue_tree = etree.fromstring(
            """
<rd:richValueData xmlns:rd="http://schemas.microsoft.com/office/spreadsheetml/2017/richdata">
  <rd:rv s="9"><rd:v>0</rd:v><rd:v>a</rd:v></rd:rv>
  <rd:rv s="0"><rd:v>0</rd:v></rd:rv>
  <rd:rv s="0"><rd:v>x</rd:v><rd:v>a</rd:v></rd:rv>
  <rd:rv s="0"><rd:v>5</rd:v><rd:v>a</rd:v></rd:rv>
  <rd:rv s="0"><rd:v>0</rd:v><rd:v>a</rd:v></rd:rv>
</rd:richValueData>
""".strip().encode("utf-8")
        )
        targets = vm_module._get_rich_data_value_targets(
            richvalue_tree=richvalue_tree,
            local_image_type_indexes={"0"},
            relationship_targets={"rId1": "xl/media/image1.png"},
        )
        self.assertEqual({"4": "xl/media/image1.png"}, targets)

    def test_map_vm_ids_to_targets_skips_invalid_metadata_rows(self):
        metadata_tree = etree.fromstring(
            """
<metadata xmlns="http://schemas.openxmlformats.org/spreadsheetml/2006/main">
  <valueMetadata>
    <bk>
      <rc />
      <rc v="9" />
      <rc v="x" />
      <rc v="0" />
    </bk>
  </valueMetadata>
</metadata>
""".strip().encode("utf-8")
        )
        with self.assertLogs(level="WARNING") as log_ctx:
            mapping = vm_module._map_vm_ids_to_targets(
                metadata_tree=metadata_tree,
                rich_data_value_targets={"x": "skip-me", "0": "xl/media/image1.png"},
            )
        self.assertEqual({"1": "xl/media/image1.png"}, mapping)
        self.assertTrue(
            any("unable to parse value metadata index 'x'" in msg for msg in log_ctx.output)
        )

    def test_get_incell_images_refs_returns_file_not_found_when_parts_are_missing(self):
        zip_buffer = self._build_archive()
        with ZipFile(zip_buffer, mode="r") as zf:
            refs, err = get_incell_images_refs(zf)
        self.assertEqual({}, refs)
        self.assertIsInstance(err, FileNotFoundError)

    def test_get_incell_images_refs_returns_xml_error_for_invalid_structure_part(self):
        zip_buffer = self._build_archive(
            rels_xml=self._valid_rels_xml(),
            structure_xml="<rd:richValueStructures",
            richvalue_xml=self._valid_richvalue_xml(),
            metadata_xml=self._valid_metadata_xml(),
        )
        with ZipFile(zip_buffer, mode="r") as zf:
            refs, err = get_incell_images_refs(zf)
        self.assertEqual({}, refs)
        self.assertIsInstance(err, etree.XMLSyntaxError)

    def test_get_incell_images_refs_uses_runtime_error_when_xml_tree_missing_without_error(self):
        zip_buffer = self._build_archive(
            rels_xml=self._valid_rels_xml(),
            structure_xml=self._valid_structure_xml(),
            richvalue_xml=self._valid_richvalue_xml(),
            metadata_xml=self._valid_metadata_xml(),
        )
        with ZipFile(zip_buffer, mode="r") as zf:
            with patch("xx2html.core.vm._get_xml_from_archive", return_value=(None, None)):
                refs, err = get_incell_images_refs(zf)
        self.assertEqual({}, refs)
        self.assertIsInstance(err, RuntimeError)
        self.assertIn("rich value structure XML", str(err))

    def test_get_incell_images_refs_uses_runtime_error_for_missing_richvalue_tree(self):
        structure_tree = etree.fromstring(self._valid_structure_xml().encode("utf-8"))
        zip_buffer = self._build_archive(
            rels_xml=self._valid_rels_xml(),
            structure_xml=self._valid_structure_xml(),
            richvalue_xml=self._valid_richvalue_xml(),
            metadata_xml=self._valid_metadata_xml(),
        )
        with ZipFile(zip_buffer, mode="r") as zf:
            with patch(
                "xx2html.core.vm._get_xml_from_archive",
                side_effect=[(structure_tree, None), (None, None)],
            ):
                refs, err = get_incell_images_refs(zf)
        self.assertEqual({}, refs)
        self.assertIsInstance(err, RuntimeError)
        self.assertIn("rich value XML", str(err))

    def test_get_incell_images_refs_uses_runtime_error_for_missing_metadata_tree(self):
        structure_tree = etree.fromstring(self._valid_structure_xml().encode("utf-8"))
        richvalue_tree = etree.fromstring(self._valid_richvalue_xml().encode("utf-8"))
        zip_buffer = self._build_archive(
            rels_xml=self._valid_rels_xml(),
            structure_xml=self._valid_structure_xml(),
            richvalue_xml=self._valid_richvalue_xml(),
            metadata_xml=self._valid_metadata_xml(),
        )
        with ZipFile(zip_buffer, mode="r") as zf:
            with patch(
                "xx2html.core.vm._get_xml_from_archive",
                side_effect=[(structure_tree, None), (richvalue_tree, None), (None, None)],
            ):
                refs, err = get_incell_images_refs(zf)
        self.assertEqual({}, refs)
        self.assertIsInstance(err, RuntimeError)
        self.assertIn("metadata XML", str(err))

    def test_get_incell_images_refs_catches_unexpected_exceptions(self):
        zip_buffer = self._build_archive(
            rels_xml=self._valid_rels_xml(),
            structure_xml=self._valid_structure_xml(),
            richvalue_xml=self._valid_richvalue_xml(),
            metadata_xml=self._valid_metadata_xml(),
        )
        with ZipFile(zip_buffer, mode="r") as zf:
            with patch(
                "xx2html.core.vm._get_relationship_targets",
                side_effect=RuntimeError("boom"),
            ):
                refs, err = get_incell_images_refs(zf)
        self.assertEqual({}, refs)
        self.assertIsInstance(err, RuntimeError)
        self.assertEqual("boom", str(err))


if __name__ == "__main__":
    unittest.main()



import unittest

from xx2html.core.links import update_links


class UpdateLinksTests(unittest.TestCase):
    def test_skips_anchors_without_href_and_rewrites_links(self):
        html = (
            '<div>'
            '<a><span>No href</span></a>'
            '<a href="#Sheet 1.A1"><span>Go local</span></a>'
            '<a href="https://example.com"><strong>Go external</strong></a>'
            '</div>'
        )

        output = update_links(html, {"Sheet 1": "sheet_001"})

        self.assertIn('about:srcdoc#sheet_001', output)
        self.assertIn('href="#sheet_001"', output)
        self.assertIn('class="xlsx_sheet-link sharepoint_visible"', output)
        self.assertIn('class="xlsx_sheet-link js_visible"', output)
        self.assertIn('<span>Go local</span>', output)

        self.assertIn('href="https://example.com"', output)
        self.assertIn('target="_blank"', output)
        self.assertIn('rel="noopener noreferrer"', output)
        self.assertIn('<strong>Go external</strong>', output)

    def test_rewrites_local_links_for_sheet_names_containing_dots(self):
        html = '<a href="#Q1.2026.A1">Go to dotted sheet</a>'

        output = update_links(html, {"Q1.2026": "sheet_abc"})

        self.assertIn('about:srcdoc#sheet_abc', output)
        self.assertIn('href="#sheet_abc"', output)
        self.assertNotIn('about:srcdoc"', output)

    def test_rewrites_local_links_for_dotted_sheet_name_without_cell_suffix(self):
        html = '<a href="#Q1.2026">Go to dotted sheet root</a>'

        output = update_links(html, {"Q1.2026": "sheet_abc"})

        self.assertIn('about:srcdoc#sheet_abc', output)
        self.assertIn('href="#sheet_abc"', output)

    def test_rewrites_excel_style_local_reference(self):
        html = '<a href="#\'Q1.2026 Sheet\'!A1">Excel style reference</a>'

        output = update_links(html, {"Q1.2026 Sheet": "sheet_q1"})

        self.assertIn('about:srcdoc#sheet_q1', output)
        self.assertIn('href="#sheet_q1"', output)

    def test_rewrites_local_links_without_cell_reference_suffix(self):
        html = '<a href="#Dashboard">Open dashboard</a>'

        output = update_links(html, {"Dashboard": "sheet_dash"})

        self.assertIn('about:srcdoc#sheet_dash', output)
        self.assertIn('href="#sheet_dash"', output)

    def test_keeps_original_local_link_when_sheet_mapping_is_missing(self):
        html = '<a href="#MissingSheet.A1">Broken ref</a>'

        output = update_links(html, {"Sheet 1": "sheet_001"})

        self.assertIn('href="#MissingSheet.A1"', output)
        self.assertNotIn("about:srcdoc", output)


if __name__ == "__main__":
    unittest.main()



from copy import deepcopy
from os import path
from pathlib import Path
import json
import logging


def initialize(
    app_config_path: str,
    config_filename: str,
    default_config: dict,
    system_config: dict,
):
    """Load app config from disk, merge defaults/system values, and return it."""
    logging.info("Application: Initializing...")

    logging.info(f"Application: Using config path |> {app_config_path}")
    Path(app_config_path).mkdir(parents=True, exist_ok=True)

    current_config = deepcopy(default_config)
    config_file_path = path.join(app_config_path, config_filename)

    if not path.exists(config_file_path):
        with open(config_file_path, "w", encoding="utf-8") as config_file:
            json.dump(current_config, config_file)
    else:
        with open(config_file_path, "r", encoding="utf-8") as config_file:
            user_config = json.load(config_file)
        if isinstance(user_config, dict):
            current_config.update(user_config)
    current_config.update(system_config)
    logging.info(f"Application: Current config |> {current_config}")
    return current_config



"""Worksheet-to-HTML table helpers."""

from collections.abc import Callable
from typing import TypedDict

from condif2css.css import CssBuilder, CssRulesRegistry
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.cell import Cell, MergedCell
from xx2html.core.types import (
    CellDimensions,
    CellRenderData,
    ColumnRenderData,
    CovaCell,
    WorksheetContents,
)
# from xx2html.core.css import CssRegistry
from xlsx2html.core import (
    rows_from_range,
    format_cell,
    unescape,
    get_cell_id,
    column_index_from_string,
    images_to_data,
    render_attrs,
    render_inline_styles,
)
import logging

COL_WIDTH__FACTOR = 17 / 2.77734375
COL_WIDTH__DEFAULT = 65  # 65px
CELL_HEIGHT__DEFAULT = 19  # 19px
# MAX_COL_REGEX = re.compile(r":(?P<max_col>[a-zA-Z]+)\d+$")
# MAX_COLS_EXCEL = 16_384


def get_worksheet_contents(
    ws: Worksheet,
    # css_registry: CssRegistry,
    # get_css_components_from_cell: Callable[[Cell | CovaCell, dict], Tuple[dict, set]],
    css_rules_registry: CssRulesRegistry,
    css_builder: CssBuilder,
    get_css_from_cell: Callable[[Cell | CovaCell | MergedCell, dict], set[str]],
    locale: None | str = None,
    ws_index: int = -1,
) -> WorksheetContents:
    """Extract normalized render data for one worksheet.

    Returns a typed payload containing rows, columns, image metadata, and
    in-cell rich-value references required by HTML rendering.
    """

    class VmCellLayoutEntry(TypedDict):
        class_name: str
        vm_id: str
        col_idx_1_based: int
        row_idx_1_based: int
        colspan: int
        rowspan: int

    merged_cell_map = {}
    used_vm_ids: set[str] = set()
    vm_ids_dimension_references: dict[str, CellDimensions] = {}
    vm_cell_vm_ids: dict[str, str] = {}
    vm_cells_layout: list[VmCellLayoutEntry] = []

    merged_cell_ranges = [cell_range.coord for cell_range in ws.merged_cells.ranges]
    excluded_cells = set(
        [
            cell
            for cell_range in merged_cell_ranges
            for rows in rows_from_range(cell_range)
            for cell in rows
        ]
    )

    for cell_range in merged_cell_ranges:
        if ":" not in str(cell_range):
            cell_range_list = list(ws[f"{cell_range}:{cell_range}"])
        else:
            cell_range_list = list(ws[cell_range])

        m_cell = cell_range_list[0][0]

        colspan = len(cell_range_list[0])
        rowspan = len(cell_range_list)
        merged_cell_map[m_cell.coordinate] = {
            "attrs": {
                "colspan": None if colspan <= 1 else colspan,
                "rowspan": None if rowspan <= 1 else rowspan,
            },
            "cells": [c for rows in cell_range_list for c in rows],
        }

        excluded_cells.remove(m_cell.coordinate)

    def get_effective_row_height(row_number: int) -> int:
        row_dim = ws.row_dimensions[row_number]
        if row_dim.hidden:
            return 0
        if row_dim.customHeight and isinstance(row_dim.height, (int, float)):
            return int(round(row_dim.height, 2))
        return CELL_HEIGHT__DEFAULT

    def process_cell(col_idx: int, cell: Cell | CovaCell | MergedCell) -> None:
        if not cell or cell.row is None:
            logging.warning("Cell without row information found, skipping processing.")
            return
        row_dim = ws.row_dimensions[cell.row]

        if cell.coordinate in excluded_cells or row_dim.hidden:
            return

        height = get_effective_row_height(cell.row)

        f_cell = None

        value = cell.value
        if isinstance(value, str):
            value = unescape(value)

        # cell_height_class = css_registry.register_height(height)
        cell_height_class = css_rules_registry.register(
            [css_builder.height(height)]
        )
        classes = set([cell_height_class])
        vm_id = None if not hasattr(cell, "_vm_id") else getattr(cell, "_vm_id")

        cell_data: CellRenderData = {  # initialization of cell_data
            "attrs": {"id": get_cell_id(cell)},
            "column": cell.column,
            "row": cell.row,
            "value": value,
            "formatted_value": format_cell(cell, locale=locale, f_cell=f_cell),
            "style": {},
            "classes": classes,
            "vm_id": vm_id,
        }

        cell_class_name = ""
        if isinstance(vm_id, str):
            cell_class_name = f"cell_{ws_index}_{col_idx}_{row_i}"

            cell_data.update({"formatted_value": ""})
            cell_data["style"].update({"position": "relative", "overflow": "hidden"})
            cell_data["classes"].update(
                [
                    f"vm-richvaluerel_rid{vm_id}",
                    cell_class_name,
                    "incell-image",
                ]
            )

            used_vm_ids.add(vm_id)

        merged_cell_info = merged_cell_map.get(cell.coordinate, {})

        if merged_cell_info:
            cell_data["attrs"].update(  # Update cell_data attrs
                merged_cell_info["attrs"]
            )

        if isinstance(vm_id, str):
            colspan = int(cell_data["attrs"].get("colspan") or 1)
            rowspan = int(cell_data["attrs"].get("rowspan") or 1)
            vm_cells_layout.append(
                {
                    "class_name": cell_class_name,
                    "vm_id": vm_id,
                    "col_idx_1_based": col_idx + 1,
                    "row_idx_1_based": cell.row,
                    "colspan": colspan,
                    "rowspan": rowspan,
                }
            )

        # new_styles, new_classes = get_css_components_from_cell(cell, merged_cell_info)
        # cell_data["style"].update(  # Update cell_data style
        #     new_styles
        # )
        new_classes = get_css_from_cell(cell, merged_cell_info)
        cell_data["classes"].update(  # Update cell_data classes
            new_classes
        )
        data_row.append(cell_data)  # Appending current cell_data to array

    columns_dimensions: dict[str, ColumnRenderData] = {}

    def first_row_process_cell(col_idx: int, cell: Cell | CovaCell | MergedCell) -> None:
        nonlocal columns_dimensions
        column_letter = get_column_letter(col_idx + 1)
        columns_dimensions[column_letter] = {
            "attrs": {},
            "index": column_letter,
            "width": COL_WIDTH__DEFAULT,
            "style": {"visibility": "visible"},
            "hidden": False,
            "collapsed": False,
        }
        process_cell(col_idx, cell)

    current_process_cell = first_row_process_cell

    data_list = []
    for row_i, row in enumerate(ws.iter_rows()):
        data_row = []
        data_list.append(data_row)
        for col_idx, cell in enumerate(row):
            current_process_cell(col_idx, cell)
        current_process_cell = process_cell

    sheet_max_column_index = len(columns_dimensions)

    for _, custom_col_dim in ws.column_dimensions.items():
        if not (
            isinstance(custom_col_dim.min, int) and isinstance(custom_col_dim.max, int)
        ):
            continue

        col_width = int(
            round(
                (custom_col_dim.width * COL_WIDTH__FACTOR)
                if custom_col_dim.customWidth
                else COL_WIDTH__DEFAULT
            )
        )

        for range_col_idx in range(
            custom_col_dim.min, min(custom_col_dim.max, sheet_max_column_index) + 1
        ):
            column_letter = get_column_letter(range_col_idx)
            if column_letter in columns_dimensions:
                column_dimensions = columns_dimensions[column_letter]

                hidden = custom_col_dim.hidden

                column_dimensions["hidden"] = hidden
                column_dimensions["collapsed"] = custom_col_dim.collapsed
                column_dimensions["style"].update(
                    {"visibility": "collapse" if (custom_col_dim.hidden) else "visible"}
                )
                column_dimensions["width"] = col_width if not hidden else 0
            else:
                logging.warning(
                    f"App (Contents): Column '{column_letter}' not found for custom dimensions!"
                )

    col_list: list[ColumnRenderData] = []
    table_width = 0

    for _, col_details in sorted(
        columns_dimensions.items(), key=lambda d: column_index_from_string(d[0])
    ):
        col_list.append(col_details)
        if not col_details["hidden"]:
            table_width += col_details["width"]

    def get_effective_col_width(col_idx_1_based: int) -> int:
        column_letter = get_column_letter(col_idx_1_based)
        col_details = columns_dimensions.get(column_letter)
        if col_details is None:
            return COL_WIDTH__DEFAULT
        width = col_details.get("width")
        if not isinstance(width, int):
            return COL_WIDTH__DEFAULT
        return max(width, 0)

    for vm_cell in vm_cells_layout:
        class_name = vm_cell["class_name"]
        vm_id = vm_cell["vm_id"]
        start_col = vm_cell["col_idx_1_based"]
        start_row = vm_cell["row_idx_1_based"]
        colspan = vm_cell["colspan"]
        rowspan = vm_cell["rowspan"]

        width_px = sum(
            get_effective_col_width(col_idx)
            for col_idx in range(start_col, start_col + colspan)
        )
        height_px = sum(
            get_effective_row_height(row_idx)
            for row_idx in range(start_row, start_row + rowspan)
        )

        if width_px <= 0:
            width_px = COL_WIDTH__DEFAULT
        if height_px <= 0:
            height_px = CELL_HEIGHT__DEFAULT

        vm_ids_dimension_references[class_name] = {
            "width": width_px,
            "height": height_px,
        }
        vm_cell_vm_ids[class_name] = vm_id

    worksheet_contents: WorksheetContents = {
        "rows": data_list,
        "cols": col_list,
        "images": images_to_data(ws),
        "vm_ids": used_vm_ids,
        "vm_ids_dimension_references": vm_ids_dimension_references,
        "vm_cell_vm_ids": vm_cell_vm_ids,
        "table_width": table_width,
    }
    return worksheet_contents


def cova_render_table(
    data: WorksheetContents,  # , append_headers, append_lineno
) -> str:
    """Render worksheet contents into a single `<table>` HTML string."""
    html = [
        "".join(
            [
                "<table  ",
                f'style="border:0; border-collapse: collapse; width: {data["table_width"]}px; table-layout: fixed;" '
                if "table_width" in data
                else 'style="border-collapse: collapse; table-layout: fixed;" ',
                " >",
            ]
        )
    ]

    sizes_row = ["<tr>"]

    html.append("<colgroup>")
    for col in data["cols"]:
        html.append(
            '<col {attrs} style="{styles}" data-value="{col_index}">'.format(
                attrs=render_attrs(col.get("attrs")),
                styles=render_inline_styles(col.get("style")),
                col_index=col["index"],
            )
        )
        width = col["width"]
        sizes_row.append(
            '<td style="width: {}px; padding: 0;"></td>'.format(width)
            if isinstance(width, int)
            else "<td></td>"
        )
    html.append("</colgroup>")
    sizes_row.append("</tr>")

    html.append("<tbody>")

    # append_headers(data, html)
    html.append(" ".join(sizes_row))

    for _, row in enumerate(data["rows"]):
        trow = ["<tr>"]
        # append_lineno(trow, i)
        for cell in row:
            images = data["images"].get((cell["column"], cell["row"])) or []
            formatted_images = []

            for img in images:
                styles = render_inline_styles(img["style"])
                img_tag = (
                    '<img width="{width}" height="{height}"'
                    'style="{styles_str}"'
                    'src="{src}"'
                    "/>"
                ).format(styles_str=styles, **img)
                formatted_images.append(img_tag)

            trow.append(
                (
                    '<td {attrs_str} style="{styles_str}" class="{classes_str}">'
                    "{formatted_images}"
                    "{formatted_value}"
                    "{incell_image}"
                    "</td>"
                ).format(
                    attrs_str=render_attrs(cell["attrs"]),
                    styles_str=render_inline_styles(cell["style"]),
                    formatted_images="\n".join(formatted_images),
                    incell_image="<img ></img>" if isinstance(cell["vm_id"], str) else "",
                    classes_str=" ".join(cell["classes"]),
                    **cell,
                )
            )

        trow.append("</tr>")
        html.append("\n".join(trow))

    html.append("</tbody>")
    html.append("</table>")
    return "\n".join(html)



import unittest

from bs4 import BeautifulSoup

from xx2html.core.cf import apply_cf_styles


class ApplyCfStylesTests(unittest.TestCase):
    def test_applies_new_classes_to_matching_cells_and_preserves_existing(self):
        html = (
            "<table><tbody>"
            '<tr><td id="Sheet1!A1" class="base">X</td></tr>'
            '<tr><td id="Sheet1!B1">Y</td></tr>'
            '<tr><td id="Sheet2!A1" class="other">Z</td></tr>'
            "</tbody></table>"
        )
        rels = [
            ("Sheet1", "A1", {"cf-red", "cf-bold"}),
            ("Sheet1", "B1", {"cf-green"}),
            ("Sheet2", "A1", {"cf-blue"}),
        ]

        output = apply_cf_styles(html, rels)

        self.assertIn('id="Sheet1!A1"', output)
        self.assertIn('class="base ', output)
        self.assertIn("cf-red", output)
        self.assertIn("cf-bold", output)

        self.assertIn('id="Sheet1!B1"', output)
        self.assertIn("cf-green", output)

        self.assertIn('id="Sheet2!A1"', output)
        self.assertIn('class="other ', output)
        self.assertIn("cf-blue", output)

    def test_ignores_non_matching_references(self):
        html = (
            "<table><tbody>"
            '<tr><td id="Sheet1!A1" class="base">X</td></tr>'
            "</tbody></table>"
        )
        rels = [
            ("Sheet1", "B2", {"cf-red"}),
            ("Unknown", "A1", {"cf-blue"}),
        ]

        output = apply_cf_styles(html, rels)

        self.assertIn('id="Sheet1!A1"', output)
        self.assertIn('class="base"', output)
        self.assertNotIn("cf-red", output)
        self.assertNotIn("cf-blue", output)

    def test_deduplicates_and_stabilizes_class_names(self):
        html = (
            "<table><tbody>"
            '<tr><td id="Sheet1!A1" class="base cf-red">X</td></tr>'
            "</tbody></table>"
        )
        rels = [("Sheet1", "A1", {"cf-red", "cf-bold"})]

        output = apply_cf_styles(html, rels)
        soup = BeautifulSoup(output, "lxml")
        classes = soup.find("td", {"id": "Sheet1!A1"}).get("class")

        self.assertEqual(["base", "cf-red", "cf-bold"], classes)


if __name__ == "__main__":
    unittest.main()



"""Extract vm-id to media target mappings from XLSX rich-data parts."""

import logging
from typing import TypeAlias
from zipfile import ZipFile

from lxml import etree
from openpyxl.packaging.relationship import get_dependents

_METADATA_XML = "xl/metadata.xml"
_RICHVALUE_REL_XML_RELS = "xl/richData/_rels/richValueRel.xml.rels"
_RICHVALUE_XML = "xl/richData/rdrichvalue.xml"
_RICHVALUES_STRUCTURE_XML = "xl/richData/rdrichvaluestructure.xml"

SHEET_MAIN_NS = "http://schemas.openxmlformats.org/spreadsheetml/2006/main"
RICHDATA_NS = "http://schemas.microsoft.com/office/spreadsheetml/2017/richdata"

EXPECTED_RICHDATA_IMAGE_TYPES = {"_localImage"}

NAMESPACES = {
    "x": SHEET_MAIN_NS,
    "rd": RICHDATA_NS,
}

VmId: TypeAlias = str
TargetPath: TypeAlias = str
RichDataValueIndex: TypeAlias = str

RelationshipTargets: TypeAlias = dict[str, TargetPath]
RichDataValueTargets: TypeAlias = dict[RichDataValueIndex, TargetPath]
VmIdToTargetMap: TypeAlias = dict[VmId, TargetPath]


def _try_parse_int(value: str | None) -> int | None:
    """Parse `value` as int, returning `None` when parsing fails."""
    if value is None:
        return None
    try:
        return int(value)
    except ValueError:
        return None


def _get_xml_from_archive(
    archive: ZipFile, file_path: str
) -> tuple[etree._Element | None, None | Exception]:
    """
    Reads an XML file from the given archive and returns it as an lxml Element.
    If the file does not exist in the archive, returns None.
    """
    try:
        with archive.open(file_path) as xml_file:
            return etree.parse(xml_file).getroot(), None
    except KeyError as key_error:
        logging.warning("File %s not found in the archive.", file_path)
        return None, key_error
    except etree.XMLSyntaxError as xml_exc:
        logging.error("XML syntax error in %s: %s", file_path, xml_exc)
        return None, xml_exc
    except Exception as exc:
        logging.error("Error reading %s from archive: %s", file_path, exc)
        return None, exc


def _get_relationship_targets(archive: ZipFile) -> RelationshipTargets:
    """Read relationship ids (`rId*`) and target paths for rich data."""
    image_rels = get_dependents(archive, _RICHVALUE_REL_XML_RELS)
    return {rel.Id: rel.Target for rel in image_rels}


def _get_local_image_type_indexes(richvalues_structure_tree: etree._Element) -> set[str]:
    """Collect rich-data structure indexes that represent local images."""
    local_image_type_indexes = set()
    for index, structure_node in enumerate(
        richvalues_structure_tree.xpath("//rd:s", namespaces=NAMESPACES)
    ):
        if structure_node.get("t") in EXPECTED_RICHDATA_IMAGE_TYPES:
            local_image_type_indexes.add(str(index))
    return local_image_type_indexes


def _get_rich_data_value_targets(
    richvalue_tree: etree._Element,
    local_image_type_indexes: set[str],
    relationship_targets: RelationshipTargets,
) -> RichDataValueTargets:
    """
    Returns a map from rich-data value index (metadata rc@v) to the target image path.
    """
    rich_data_value_targets: RichDataValueTargets = {}
    for value_index, rich_value_node in enumerate(
        richvalue_tree.xpath("//rd:rv", namespaces=NAMESPACES)
    ):
        rich_value_type_index = rich_value_node.get("s")
        if rich_value_type_index not in local_image_type_indexes:
            continue

        # Keep compatibility with the expected structure the old parser required.
        if len(rich_value_node) != 2:
            continue

        rel_index = _try_parse_int(rich_value_node[0].text)
        if rel_index is None:
            continue

        target_path = relationship_targets.get(f"rId{rel_index + 1}")
        if target_path is None:
            continue
        rich_data_value_targets[str(value_index)] = target_path

    return rich_data_value_targets


def _map_vm_ids_to_targets(
    metadata_tree: etree._Element, rich_data_value_targets: RichDataValueTargets
) -> VmIdToTargetMap:
    """
    Returns a map from vm_id (worksheet cell vm attr) to target image path.
    """
    vm_id_to_target: VmIdToTargetMap = {}
    for record in metadata_tree.xpath("//x:valueMetadata/x:bk/x:rc", namespaces=NAMESPACES):
        rich_data_value_index = record.get("v")
        if rich_data_value_index is None:
            continue

        target_path = rich_data_value_targets.get(rich_data_value_index)
        if target_path is None:
            continue

        vm_id_as_int = _try_parse_int(rich_data_value_index)
        if vm_id_as_int is None:
            logging.warning(
                "get_incell_images_refs: unable to parse value metadata index '%s'.",
                rich_data_value_index,
            )
            continue

        vm_id_to_target[str(vm_id_as_int + 1)] = target_path

    return vm_id_to_target


def get_incell_images_refs(
    archive: ZipFile,
) -> tuple[VmIdToTargetMap, Exception | None]:
    """
    Extracts in-cell image references from the given archive and returns:
    - vm_id -> target image path
    """
    archive_namelist = archive.namelist()
    required_files = (
        _RICHVALUE_REL_XML_RELS,
        _METADATA_XML,
        _RICHVALUE_XML,
        _RICHVALUES_STRUCTURE_XML,
    )
    if not all(required_file in archive_namelist for required_file in required_files):
        return {}, FileNotFoundError(
            "Missing required files in the archive for in-cell images extraction."
        )

    try:
        relationship_targets = _get_relationship_targets(archive)
        logging.info(
            "get_incell_images_refs: relationship_targets -> %s", relationship_targets
        )

        richvalues_structure_tree, xml_error = _get_xml_from_archive(
            archive, _RICHVALUES_STRUCTURE_XML
        )
        if xml_error is not None or richvalues_structure_tree is None:
            return {}, xml_error or RuntimeError("Failed to read rich value structure XML.")
        local_image_type_indexes = _get_local_image_type_indexes(richvalues_structure_tree)

        richvalue_tree, xml_error = _get_xml_from_archive(archive, _RICHVALUE_XML)
        if xml_error is not None or richvalue_tree is None:
            return {}, xml_error or RuntimeError("Failed to read rich value XML.")
        rich_data_value_targets = _get_rich_data_value_targets(
            richvalue_tree, local_image_type_indexes, relationship_targets
        )
        logging.info(
            "get_incell_images_refs: rich_data_value_targets -> %s",
            rich_data_value_targets,
        )

        metadata_tree, xml_error = _get_xml_from_archive(archive, _METADATA_XML)
        if xml_error is not None or metadata_tree is None:
            return {}, xml_error or RuntimeError("Failed to read metadata XML.")
        vm_id_to_target = _map_vm_ids_to_targets(metadata_tree, rich_data_value_targets)
        logging.info("get_incell_images_refs: vm_id_to_target -> %s", vm_id_to_target)

        return vm_id_to_target, None
    except Exception as incell_exc:
        logging.warning(
            "get_incell_images_refs: unable to read in-cell images due to: %r",
            incell_exc,
        )
        return {}, incell_exc



import unittest
import warnings
from types import SimpleNamespace
from xml.etree.ElementTree import Element, SubElement
from unittest.mock import patch

from openpyxl import Workbook

import xx2html.core.patches.openpyxl as patch_module
from xx2html.core.patches.openpyxl import apply_patches, cova_bind_cells, cova_parse_cell
from xx2html.core.types import CovaCell


def _make_parser(
    *,
    data_only: bool = True,
    rich_text: bool = False,
    date_formats: set[int] | None = None,
):
    return SimpleNamespace(
        data_only=data_only,
        rich_text=rich_text,
        col_counter=0,
        row_counter=1,
        date_formats=date_formats or set(),
        timedelta_formats=set(),
        epoch=None,
        shared_strings=["zero", "one", "two"],
        parse_formula=lambda _element: "FORMULA_RESULT",
    )


def _make_cell_element(
    *,
    data_type: str = "n",
    coordinate: str = "A1",
    style_id: str = "0",
    value: str | None = None,
    include_formula: bool = False,
    include_inline: bool = False,
) -> Element:
    element = Element("c", {"t": data_type, "r": coordinate, "s": style_id})
    if value is not None:
        value_tag = SubElement(element, patch_module.VALUE_TAG)
        value_tag.text = value
    if include_formula:
        SubElement(element, patch_module.FORMULA_TAG)
    if include_inline:
        SubElement(element, patch_module.INLINE_STRING)
    return element


class CovaParseCellTests(unittest.TestCase):
    def test_formula_cells_use_parse_formula_when_data_only_is_false(self):
        parser = _make_parser(data_only=False)
        cell_element = _make_cell_element(
            data_type="n", value="5", include_formula=True
        )

        parsed = cova_parse_cell(parser, cell_element)

        self.assertEqual("f", parsed["data_type"])
        self.assertEqual("FORMULA_RESULT", parsed["value"])

    def test_date_overflow_marks_cell_as_error_and_warns(self):
        parser = _make_parser(date_formats={1})
        cell_element = _make_cell_element(data_type="n", style_id="1", value="42")

        with patch(
            "xx2html.core.patches.openpyxl.from_excel", side_effect=OverflowError
        ):
            with warnings.catch_warnings(record=True) as warning_records:
                warnings.simplefilter("always")
                parsed = cova_parse_cell(parser, cell_element)

        self.assertEqual("e", parsed["data_type"])
        self.assertEqual("#VALUE!", parsed["value"])
        self.assertTrue(
            any("outside the limits for dates" in str(w.message) for w in warning_records)
        )

    def test_inline_rich_text_uses_parse_richtext_string(self):
        parser = _make_parser(rich_text=True)
        cell_element = _make_cell_element(data_type="inlineStr", include_inline=True)

        with patch(
            "xx2html.core.patches.openpyxl.parse_richtext_string",
            return_value="rich-text-value",
        ) as rich_text_mock:
            parsed = cova_parse_cell(parser, cell_element)

        rich_text_mock.assert_called_once()
        self.assertEqual("s", parsed["data_type"])
        self.assertEqual("rich-text-value", parsed["value"])

    def test_inline_plain_text_uses_text_from_tree_content(self):
        parser = _make_parser(rich_text=False)
        cell_element = _make_cell_element(data_type="inlineStr", include_inline=True)
        text_obj = SimpleNamespace(content="plain-text")

        with patch(
            "xx2html.core.patches.openpyxl.Text.from_tree", return_value=text_obj
        ) as text_mock:
            parsed = cova_parse_cell(parser, cell_element)

        text_mock.assert_called_once()
        self.assertEqual("s", parsed["data_type"])
        self.assertEqual("plain-text", parsed["value"])


class CovaBindCellsTests(unittest.TestCase):
    def test_bind_cells_creates_covacell_and_updates_current_row(self):
        workbook = Workbook()
        worksheet = workbook.active
        parser = SimpleNamespace(
            parse=lambda: iter(
                [
                    (
                        1,
                        [
                            {
                                "style_id": 0,
                                "row": 1,
                                "column": 1,
                                "value": "hello",
                                "data_type": "s",
                                "vm_id": "9",
                            }
                        ],
                    )
                ]
            )
        )
        reader = SimpleNamespace(parser=parser, ws=worksheet)

        cova_bind_cells(reader)

        bound_cell = worksheet._cells[(1, 1)]
        self.assertIsInstance(bound_cell, CovaCell)
        self.assertEqual("hello", bound_cell._value)
        self.assertEqual("s", bound_cell.data_type)
        self.assertEqual("9", bound_cell._vm_id)
        self.assertEqual(1, worksheet._current_row)

    def test_apply_patches_registers_cova_hooks(self):
        apply_patches()
        self.assertIs(patch_module.WorkSheetParser.parse_cell, patch_module.cova_parse_cell)
        self.assertIs(
            patch_module.WorksheetReader.bind_cells, patch_module.cova_bind_cells
        )


if __name__ == "__main__":
    unittest.main()



from typing import Callable, Literal, Tuple
from openpyxl.cell import Cell
from openpyxl.styles.colors import Color

from xx2html.core.types import CovaCell
from condif2css.color import argb_to_css

import logging

DEFAULT_BORDER_STYLE = (
    "border-{direction}-style: solid; border-{direction}-width: 1px; "
)
BORDER_STYLES = {
    "dashDot": None,
    "dashDotDot": None,
    "dashed": "border-{direction}-style: dashed; ",
    "dotted": "border-{direction}-style: dotted; ",
    "double": "border-{direction}-style: double; ",
    "hair": None,
    "medium": "border-{direction}-style: solid; border-{direction}-width: 2px; ",
    "mediumDashDot": "border-{direction}-style: solid; border-{direction}-width: 2px; ",
    "mediumDashDotDot": "border-{direction}-style: solid; border-{direction}-width: 2px; ",
    "mediumDashed": "border-{direction}-style: dashed; border-{direction}-width: 2px; ",
    "slantDashDot": None,
    "thick": "border-{direction}-style: solid; border-{direction}-width: 1px;",
    "thin": "border-{direction}-style: solid; border-{direction}-width: 1px; ",
}


class CssRegistry:
    """Register reusable CSS class names generated from cell style fragments."""

    def __init__(
        self,
        get_css_color: Callable[[Color], str | None],
        classes: dict[str, object] | None = None,
    ) -> None:
        self.get_css_color = get_css_color
        self.classes = {} if classes is None else classes

    def register_font_size(self, size: int) -> str:
        class_name = f"xlsx_cell_font_size_{size}"
        if class_name not in self.classes:
            self.classes[class_name] = f"font-size: {size}px;"
        return class_name

    def register_height(self, size: int) -> str:
        class_name = f"xlsx_element_height_{size}"
        if class_name not in self.classes:
            self.classes[class_name] = f"height: {size}px;"
        return class_name

    def register_font_color(self, color: Color) -> str | None:
        aRGB = self.get_css_color(color)
        if aRGB is not None:
            class_name = f"xlsx_cell_font_color_{aRGB}"

            if class_name not in self.classes:
                self.classes[class_name] = f"color: {argb_to_css(aRGB)};"
            return class_name
        else:
            return None

    def register_background_color(self, color: Color) -> str | None:
        aRGB = self.get_css_color(color)
        if aRGB is not None:
            class_name = f"xlsx_cell_background_color_{aRGB}"

            if class_name not in self.classes:
                self.classes[class_name] = f"background-color : {argb_to_css(aRGB)}"

            return class_name
        else:
            return None

    def register_font_underline(self) -> str:
        class_name = "xlsx_cell_font_underline"
        if class_name not in self.classes:
            self.classes[class_name] = "text-decoration: underline;"
        return class_name

    def register_font_bold(self) -> str:
        class_name = "xlsx_cell_font_bold"
        if class_name not in self.classes:
            self.classes[class_name] = "font-weight: bold;"
        return class_name

    def register_font_italic(self) -> str:
        class_name = "xlsx_cell_font_italic"
        if class_name not in self.classes:
            self.classes[class_name] = "font-style: italic;"
        return class_name

    def register_border(
        self,
        style: str | None,
        direction: Literal["right", "left", "top", "bottom"],
        color: Color,
    ) -> str | None:
        if style is None:
            return None

        border_style = BORDER_STYLES.get(style)
        border_style = DEFAULT_BORDER_STYLE if border_style is None else border_style

        class_name = f"border_{style}_{direction[0:1]}"

        class_value = border_style.format(direction=direction)

        aRGB = self.get_css_color(color)

        if aRGB is not None:
            class_name = f"{class_name}_{aRGB}"
            class_value = (
                f"{class_value} border-{direction}-color: {argb_to_css(aRGB)};"
            )

        if class_name not in self.classes:
            self.classes[class_name] = class_value

        return class_name


def get_border_classes_from_cell(cell: Cell | CovaCell, css_registry: CssRegistry):
    """Return CSS border classes required to render a cell border."""
    classes = set()

    for b_dir in ["right", "left", "top", "bottom"]:
        b_s = getattr(cell.border, b_dir)
        if not b_s:
            continue

        border_class = css_registry.register_border(
            b_s.style,
            direction=b_dir,  # type:ignore
            color=b_s.color,
        )
        if border_class is not None:
            classes.add(border_class)

    return classes


def create_get_css_components_from_cell(css_registry: CssRegistry):
    """Build a function that extracts inline styles and class names from a cell."""

    def get_css_components_from_cell(
        cell: Cell | CovaCell,
        merged_cell_map=None,  # odefault_cell_border="none"
    ) -> Tuple[dict, set]:
        """Return `(styles, classes)` for a single cell (including merged edges)."""
        merged_cell_map = merged_cell_map or {}

        classes = set()

        # h_styles = {"border-collapse": "collapse"}
        h_styles = {}
        classes.update(get_border_classes_from_cell(cell, css_registry))
        if merged_cell_map:
            # TODO edged_cells
            for m_cell in merged_cell_map["cells"]:
                classes.update(get_border_classes_from_cell(m_cell, css_registry))

        # for b_dir in ["border-right", "border-left", "border-top", "border-bottom"]:
        # style_tag = b_dir + "-style"
        # if (b_dir not in b_styles) and (style_tag not in b_styles):
        # b_styles[b_dir] = default_cell_border
        # h_styles.update(b_styles)

        if cell.alignment.horizontal:
            h_styles["text-align"] = cell.alignment.horizontal
        if cell.alignment.vertical:
            h_styles["vertical-align"] = cell.alignment.vertical

        # with contextlib.suppress(AttributeError):
        if hasattr(cell, "fill") and hasattr(cell.fill, "patternType"):
            patternType = cell.fill.patternType
            if patternType == "solid":
                background_color_class = css_registry.register_background_color(
                    cell.fill.fgColor
                )
                if background_color_class is not None:
                    classes.add((background_color_class))
                # h_styles["background-color"] = get_css_color(cell.fill.fgColor)
            elif patternType is not None:
                # TODO patternType != 'solid'
                 logging.warning(
                    f"css (components): Pattern type is not supported: {cell.fill.patternType}"
                )

        if cell.font:
            classes.add(css_registry.register_font_size(int(cell.font.sz)))
            # h_styles["font-size"] = "%spx" % cell.font.sz
            if cell.font.color:
                font_color_class = css_registry.register_font_color(cell.font.color)
                if font_color_class is not None:
                    classes.add(font_color_class)
                # h_styles["color"] = get_css_color(cell.font.color)
            if cell.font.b:
                classes.add(css_registry.register_font_bold())
                # h_styles["font-weight"] = "bold"
            if cell.font.i:
                classes.add(css_registry.register_font_italic())
                # h_styles["font-style"] = "italic"
            if cell.font.u:
                classes.add(css_registry.register_font_underline())
                # h_styles["font-decoration"] = "underline"

        return h_styles, classes

    return get_css_components_from_cell



"""Rewrite worksheet and external links in generated HTML."""

from copy import deepcopy
from bs4 import BeautifulSoup


def update_links(
    html: str,
    encoded_sheet_names: dict[str, str],  # sheet names
    update_local_links: bool = True,
    update_ext_links: bool = True,
) -> str:
    """Rewrite anchor tags for worksheet-local and external navigation."""
    soup = BeautifulSoup(html, "lxml")

    def resolve_sheet_name(local_reference: str) -> str:
        """Resolve local reference text to a worksheet name key."""
        if local_reference in encoded_sheet_names:
            return local_reference

        if "!" in local_reference:
            sheet_name = local_reference.split("!", 1)[0]
            if sheet_name.startswith("'") and sheet_name.endswith("'"):
                return sheet_name[1:-1].replace("''", "'")
            return sheet_name

        if "." in local_reference:
            candidate = local_reference.rsplit(".", 1)[0]
            if candidate in encoded_sheet_names:
                return candidate

        return local_reference

    for anchor_tag in soup.find_all("a"):
        if "class" in anchor_tag.attrs and "xlsx_sheet-link" in anchor_tag["class"]:
            continue

        href = anchor_tag.get("href")
        if not isinstance(href, str) or href == "":
            continue

        is_local_anchor = href.startswith("#")

        if is_local_anchor:
            if update_local_links:
                local_reference = href[1:]
                sheet_name = resolve_sheet_name(local_reference)
                enc_sheet_name = encoded_sheet_names.get(sheet_name)
                if enc_sheet_name is None:
                    continue
                # sheet_name = ''.join(e for e in possible_sheet_name if e.isalnum())

                # final_href = f'#sheet-{sheet_name}' if sheet_name in usable_names else ''
                final_href = f"#{enc_sheet_name}"

                sharepoint_anchor_tag = soup.new_tag(
                    "a",
                    attrs={
                        "href": f"about:srcdoc{final_href}",
                        "class": "xlsx_sheet-link sharepoint_visible",
                    },
                )

                js_anchor_tag = soup.new_tag(
                    "a",
                    attrs={
                        "href": final_href,
                        "class": "xlsx_sheet-link js_visible",
                        # , 'target': '_blank'
                    },
                )

                for child in anchor_tag.contents:
                    sharepoint_anchor_tag.append(deepcopy(child))
                    js_anchor_tag.append(deepcopy(child))

                # print(f"sharepoint: {sharepoint_anchor_tag}")
                # print(f"js: {js_anchor_tag}")

                anchor_tag.replace_with(sharepoint_anchor_tag, js_anchor_tag)
        else:
            if update_ext_links:
                external_anchor_tag = soup.new_tag(
                    "a",
                    attrs={
                        "href": href,
                        "class": "xlsx_sheet-link js_visible",
                        "target": "_blank",
                        "rel": "noopener noreferrer",
                    },
                )
                for child in anchor_tag.contents:
                    external_anchor_tag.append(deepcopy(child))
                anchor_tag.replace_with(external_anchor_tag)
    return str(soup)



import tempfile
import unittest
from pathlib import Path

from xx2html import create_xlsx_transform


FIXTURES_DIR = Path(__file__).resolve().parent / "fixtures"

SHEET_HTML = (
    '<section id="{enc_sheet_name}" data-sheet="{sheet_name}">'
    "{table_generated_html}"
    "</section>"
)
SHEETNAME_HTML = '<a class="sheet-nav" href="#{sheet_name}.A1">{sheet_name}</a>'
INDEX_HTML = (
    "<!doctype html><html><head>"
    "{fonts_html}{core_css_html}{user_css_html}{generated_css_html}"
    "{generated_incell_css_html}{conditional_css_html}"
    "</head><body>{sheets_names_generated_html}{sheets_generated_html}{safari_js}</body></html>"
)


def _build_transform():
    return create_xlsx_transform(
        sheet_html=SHEET_HTML,
        sheetname_html=SHEETNAME_HTML,
        index_html=INDEX_HTML,
        fonts_html="",
        core_css="",
        user_css="",
        safari_js="",
        apply_cf=False,
    )


class TransformFailureTests(unittest.TestCase):
    def test_missing_source_file_returns_false_and_error(self):
        transform = _build_transform()
        missing_source = "__missing__/missing.xlsx"

        with tempfile.TemporaryDirectory() as tmp_dir:
            output_file = Path(tmp_dir) / "out.html"
            ok, err = transform(missing_source, str(output_file), "en_US")

        self.assertFalse(ok)
        self.assertIsInstance(err, str)
        self.assertIn("FileNotFoundError", err)

    def test_destination_directory_returns_false_and_error(self):
        transform = _build_transform()
        source_file = FIXTURES_DIR / "merged_cells_cf.xlsx"

        with tempfile.TemporaryDirectory() as tmp_dir:
            ok, err = transform(str(source_file), tmp_dir, "en_US")

        self.assertFalse(ok)
        self.assertIsInstance(err, str)
        self.assertIn("IsADirectoryError", err)


if __name__ == "__main__":
    unittest.main()



"""Shared type aliases and typed payload models used in core transforms."""

from collections.abc import Callable
from typing import TypeAlias, TypedDict

from openpyxl.cell import Cell


CellCoordinate: TypeAlias = tuple[int | str, int]
TransformResult: TypeAlias = tuple[bool, str | None]
XlsxTransformCallable: TypeAlias = Callable[[str, str, str], TransformResult]
ConditionalFormattingRelation: TypeAlias = tuple[str, str, set[str]]


class ImageRenderData(TypedDict):
    """Image payload used while rendering worksheet cell attachments."""

    width: int
    height: int
    style: dict[str, str]
    src: str


class CellDimensions(TypedDict):
    """Pixel dimensions for a rendered cell box."""

    width: int
    height: int


class CellRenderData(TypedDict):
    """Normalized render model for one `<td>` element."""

    attrs: dict[str, object]
    column: int | str
    row: int
    value: object
    formatted_value: str
    style: dict[str, str]
    classes: set[str]
    vm_id: str | None


class ColumnRenderData(TypedDict):
    """Normalized render model for one `<col>` element."""

    attrs: dict[str, object]
    index: str
    width: int
    style: dict[str, str]
    hidden: bool
    collapsed: bool


class WorksheetContents(TypedDict):
    """Aggregate render payload for one worksheet table."""

    rows: list[list[CellRenderData]]
    cols: list[ColumnRenderData]
    images: dict[CellCoordinate, list[ImageRenderData]]
    vm_ids: set[str]
    vm_ids_dimension_references: dict[str, CellDimensions]
    vm_cell_vm_ids: dict[str, str]
    table_width: int


class CovaCell(Cell):
    """Openpyxl cell extension that stores rich-value metadata (`vm_id`)."""

    _vm_id = None

    def __init__(
        self, worksheet, row, column, value=None, style_array=None, vm_id=None
    ):
        super().__init__(worksheet, row, column, value, style_array)
        # if vm_id is not None:
        if isinstance(vm_id, str):
            self._vm_id = vm_id

    def __repr__(self):
        if hasattr(self, "_vm_id"):
            # if self._vm_id > 0:
            if isinstance(self._vm_id, str):
                return "<CovaCell {0!r}.{1} vm_id:{2}>".format(
                    self.parent.title, self.coordinate, self._vm_id
                )
        return "<CovaCell {0!r}.{1}>".format(self.parent.title, self.coordinate)



# Copyright (c) 2024-2026 gonzalo covarrubias <gocova.dev+xx2html@gmail.com>

"""Monkey patches required to expose rich-value metadata in openpyxl cells."""

# Monkey patch for: WorkSheetParser
from openpyxl.worksheet._reader import WorkSheetParser

# Monkey patch for: WorksheetReader
from openpyxl.worksheet._reader import WorksheetReader

## required imports
from warnings import warn
from openpyxl.utils import coordinate_to_tuple
from openpyxl.utils.datetime import from_excel, from_ISO8601
from openpyxl.worksheet._reader import (
    parse_richtext_string,
    VALUE_TAG, FORMULA_TAG, INLINE_STRING,
    _cast_number, # type: ignore
)
from openpyxl.cell.text import Text

from xx2html.core.types import CovaCell


def cova_parse_cell(self, element) -> dict[str, object]:
    """
    Parse a cell element into a dictionary containing the cell's row, column, value, data type, style id, and vm id.

    :param element: The cell element to parse
    :return: A dictionary containing the cell's row, column, value, data type, style id, and vm id
    """
    data_type = element.get("t", "n")
    coordinate = element.get("r")
    style_id = element.get("s", 0)
    vm_id = element.get("vm", None)

    if style_id:
        style_id = int(style_id)

    # if vm_id:
    #     vm_id = int(vm_id)

    if data_type == "inlineStr":
        value = None
    else:
        value = element.findtext(VALUE_TAG, None) or None

    if coordinate:
        row, column = coordinate_to_tuple(coordinate)
        self.col_counter = column
    else:
        self.col_counter += 1
        row, column = self.row_counter, self.col_counter

    if not self.data_only and element.find(FORMULA_TAG) is not None:
        data_type = "f"
        value = self.parse_formula(element)

    elif value is not None:
        if data_type == "n":
            value = _cast_number(value)
            if style_id in self.date_formats:
                data_type = "d"
                try:
                    value = from_excel(
                        value, self.epoch, timedelta=style_id in self.timedelta_formats
                    )
                except (OverflowError, ValueError):
                    msg = f"""Cell {coordinate} is marked as a date but the serial value {value} is outside the limits for dates. The cell will be treated as an error."""
                    warn(msg)
                    data_type = "e"
                    value = "#VALUE!"
        elif data_type == "s":
            value = self.shared_strings[int(value)]
        elif data_type == "b":
            value = bool(int(value))
        elif data_type == "str":
            data_type = "s"
        elif data_type == "d":
            value = from_ISO8601(value)

    elif data_type == "inlineStr":
        child = element.find(INLINE_STRING)
        if child is not None:
            data_type = "s"
            if self.rich_text:
                value = parse_richtext_string(child)
            else:
                text_child = Text.from_tree(child)
                value = text_child.content if text_child else ""

    return {
        "row": row,
        "column": column,
        "value": value,
        "data_type": data_type,
        "style_id": style_id,
        "vm_id": vm_id,
    }


def cova_bind_cells(self) -> None:
    """
    Bind CovaCells to the worksheet.

    This method is used to process the parsed data and bind it to the worksheet.
    It creates a CovaCell for each cell in the parsed data and sets the value and data_type.
    The CovaCells are then stored in the worksheet's _cells dictionary.

    """
    for idx, row in self.parser.parse():
        for cell in row:
            style = self.ws.parent._cell_styles[cell["style_id"]]
            cova_cell = CovaCell(
                self.ws,
                row=cell["row"],
                column=cell["column"],
                style_array=style,
                vm_id=cell["vm_id"],
            )
            cova_cell._value = cell["value"]
            cova_cell.data_type = cell["data_type"]

            self.ws._cells[(cell["row"], cell["column"])] = cova_cell

    if self.ws._cells:
        self.ws._current_row = self.ws.max_row  # use cells not row dimensions



def apply_patches() -> None:
    """Install parser/reader monkey patches used by `xx2html`."""
    WorkSheetParser.parse_cell = cova_parse_cell
    WorksheetReader.bind_cells = cova_bind_cells



"""Conditional-formatting HTML post-processing."""

import logging
from bs4 import BeautifulSoup

from xx2html.core.types import ConditionalFormattingRelation


def apply_cf_styles(
    html: str, cf_style_relations: list[ConditionalFormattingRelation]
) -> str:
    """Attach generated conditional-formatting class names to target cells."""
    soup = BeautifulSoup(html, "lxml")
    for sheet_name, cell_ref, class_names in cf_style_relations:
        class_names_str = " ".join(class_names)
        logging.debug(
            f"apply_cf_styles: '{sheet_name}!{cell_ref}' with additional class_names: {class_names_str}"
        )
        for cell_tag in soup.find_all("td", id=f"{sheet_name}!{cell_ref}"):
            previous_classes = cell_tag.get("class")
            previous_classes = (
                previous_classes if previous_classes is not None else []
            )
            for class_name in sorted(class_names):
                if class_name not in previous_classes:
                    previous_classes.append(class_name)
            cell_tag["class"] = previous_classes
    return str(soup)



"""Core transformation API for converting XLSX workbooks to HTML."""

import logging
from zipfile import ZipFile

from openpyxl import load_workbook
from openpyxl.styles.differential import DifferentialStyleList

# Monkey patch!
from xx2html.core.cf import apply_cf_styles
from xx2html.core.patches.openpyxl import apply_patches

from .incell import get_incell_css
from .links import update_links
from .types import (
    ConditionalFormattingRelation,
    TransformResult,
    XlsxTransformCallable,
)
from .utils import cova_render_table, get_worksheet_contents
from .vm import get_incell_images_refs

# from .css import CssRegistry, create_get_css_components_from_cell
from condif2css.processor import process_conditional_formatting
from condif2css.themes import get_theme_colors
from condif2css.core import create_themed_css_color_resolver
from condif2css.color import argb_to_css
from condif2css.css import CssBuilder, CssRulesRegistry, create_get_css_from_cell

_PATCHES_APPLIED = False


def apply_openpyxl_patches() -> None:
    """Apply required openpyxl monkey patches once per process."""
    global _PATCHES_APPLIED
    if _PATCHES_APPLIED:
        return
    logging.debug("xx2html: applying required openpyxl monkey patches")
    apply_patches()
    _PATCHES_APPLIED = True


# Explicitly apply monkey patches required for vm-aware parsing.
apply_openpyxl_patches()


def create_xlsx_transform(
    sheet_html: str,
    sheetname_html: str,
    index_html: str,
    fonts_html: str,
    core_css: str,
    user_css: str,
    safari_js: str,
    update_local_links: bool = True,
    # prepare_iframe_noscript: bool = True,
    apply_cf: bool = False,
    fail_ok: bool = True,
) -> XlsxTransformCallable:
    """Build and return a configured XLSX-to-HTML transform function.

    The returned callable takes `(source, dest, locale)` and writes a complete
    HTML document to `dest`.
    """

    def transform_xlsx(
        source: str, dest: str, locale: str
    ) -> TransformResult:
        """Transform one XLSX file into one HTML file."""
        workbook = None
        workbook_archive = None
        try:
            logging.info(f"Transform (out): Opening '{dest}' for writing...")
            with open(dest, "w", encoding="utf-8") as output:
                sheet_navigation_links = []
                sheet_html_sections = []

                logging.info(f"Transform (wb): Reading '{source}' as xlsx file...")
                workbook = load_workbook(source, data_only=True, rich_text=True)

                logging.debug("Transform (wb|css): Reading theme colors...")
                theme_argb_palette = get_theme_colors(workbook)
                pre_get_css_color = create_themed_css_color_resolver(theme_argb_palette)

                def get_css_color(color):
                    argb_color = pre_get_css_color(color)
                    if argb_color is None:
                        return None
                    return argb_to_css(argb_color)

                css_builder = CssBuilder(get_css_color)
                css_registry = CssRulesRegistry()
                get_css_from_cell = create_get_css_from_cell(
                    css_registry, css_builder=css_builder
                )

                css_cf_registry = CssRulesRegistry(prefix="xx2h_cf")
                get_cf_css_from_diff = create_get_css_from_cell(
                    css_registry=css_cf_registry, css_builder=css_builder
                )

                logging.debug("Transform (wb|incell): Reading incell images...")
                incell_images_refs = {}
                try:
                    workbook_archive = ZipFile(source, "r")
                    incell_images_refs, incell_error = get_incell_images_refs(
                        workbook_archive
                    )
                    if incell_error is not None:
                        raise incell_error
                    logging.info("Transform (wb|incell): Reading complete!")
                except Exception as incell_exc:
                    logging.warning(
                        "Transform (wb|incell): Unable to read incell images due to: %r",
                        incell_exc,
                    )
                    if workbook_archive is not None:
                        workbook_archive.close()
                        workbook_archive = None

                vm_ids = set()
                vm_ids_dimension_references = dict()
                vm_cell_vm_ids = dict()

                encoded_sheet_names = dict()
                sheet_names = workbook.sheetnames
                conditional_formatting_rule_details = {}

                for sheet_name in sheet_names:
                    worksheet = workbook[sheet_name]

                    worksheet_index = workbook.index(worksheet)
                    encoded_sheet_name = f"sheet_{hex(worksheet_index)[2:].zfill(3)}"

                    encoded_sheet_names[sheet_name] = encoded_sheet_name

                    if worksheet.sheet_state == "visible":
                        logging.info(
                            f"Application (ws): Sheet[{worksheet_index}]:'{sheet_name}' (enc_sheet_name: {encoded_sheet_name}) -> is visible"
                        )

                        contents = get_worksheet_contents(
                            worksheet,
                            css_rules_registry=css_registry,
                            css_builder=css_builder,
                            get_css_from_cell=get_css_from_cell,
                            locale=locale,
                            ws_index=worksheet_index,
                        )

                        logging.info(
                            f" {encoded_sheet_name} --> vm_ids: {contents['vm_ids']}"
                        )
                        vm_ids.update(contents["vm_ids"])
                        vm_ids_dimension_references.update(
                            contents["vm_ids_dimension_references"]
                        )
                        vm_cell_vm_ids.update(contents["vm_cell_vm_ids"])

                        sheet_html_sections.append(
                            sheet_html.format(
                                enc_sheet_name=encoded_sheet_name,
                                sheet_name=sheet_name,
                                table_generated_html=cova_render_table(contents),
                            )
                        )

                        sheet_navigation_links.append(
                            sheetname_html.format(
                                enc_sheet_name=encoded_sheet_name, sheet_name=sheet_name
                            )
                        )

                        if apply_cf:
                            logging.info(
                                f"Application (wb|cf): Processing conditional formatting for '{sheet_name}'"
                            )
                            conditional_formatting_rule_details.update(
                                process_conditional_formatting(
                                    worksheet, fail_ok=fail_ok
                                )
                            )

                generated_css = "\n".join(css_registry.get_rules())

                if workbook_archive is not None:
                    logging.debug(
                        "Transform (wb|incell): Preparing incell images output..."
                    )
                    generated_incell_css = get_incell_css(
                        vm_ids,
                        vm_ids_dimension_references,
                        vm_cell_vm_ids,
                        incell_images_refs,
                        workbook_archive,
                    )
                else:
                    generated_incell_css = ""

                logging.info(
                    f"Transform (html|1): Pass 1 --> Preparing {len(conditional_formatting_rule_details)} conditional formatting styles..."
                )
                cf_style_relations: list[ConditionalFormattingRelation] = []
                if hasattr(workbook, "_differential_styles") and isinstance(
                    workbook._differential_styles,  # type: ignore
                    DifferentialStyleList,
                ):
                    for _, details in conditional_formatting_rule_details.items():
                        sheet_name, cell_ref, _, dxf_id, _ = details
                        class_names = get_cf_css_from_diff(
                            workbook._differential_styles[dxf_id],  # type: ignore
                            is_important=True,
                        )
                        cf_style_relations.append((sheet_name, cell_ref, class_names))
                logging.debug(
                    f"Transform: Resulting conditional formatting styles: {cf_style_relations}"
                )

                logging.info("Transform (html|2): Pass 2 --> Preparing html")
                css_rules = "\n".join(css_cf_registry.get_rules())
                html = (
                    index_html.format(
                        sheets_generated_html="\n".join(sheet_html_sections),
                        sheets_names_generated_html="\n".join(sheet_navigation_links),
                        source_filename=source,
                        fonts_html=fonts_html,
                        core_css_html=f"<style>{core_css}</style>",
                        user_css_html=f"<style>{user_css}</style>",
                        generated_css_html=f"<style>{generated_css}</style>",
                        generated_incell_css_html=f"<style>{generated_incell_css}</style>",
                        safari_js=f"<script>{safari_js}</script>",
                        conditional_css_html=f"<style>/*conditional formatting*/\n{css_rules}</style>",
                    )
                    .replace('"$"', "$")
                    .replace('"-"', "-")
                )

                logging.info("Transform (html|3): Pass 3 --> Updating links...")
                html_with_updated_links = update_links(
                    html, encoded_sheet_names, update_local_links=update_local_links
                )

                logging.info(
                    "Transform (html|4): Pass 4 --> Applying conditional formatting styles..."
                )
                html_with_cf_styles = apply_cf_styles(
                    html_with_updated_links, cf_style_relations
                )

                logging.info(f"Transform (out): Writing output: {dest}")
                output.write(html_with_cf_styles)

            logging.info("Transform: Done!")
            return (True, None)
        except Exception as exc:
            logging.exception("Transform failed for '%s' -> '%s'", source, dest)
            return (False, repr(exc))
        finally:
            if workbook_archive is not None:
                logging.info("Transform (wb|incell): Closing archive...")
                workbook_archive.close()
            if workbook is not None:
                logging.info(f"Transform (wb): Closing wb: {source}")
                workbook.close()

    return transform_xlsx
